"""utils module, part of cuisto.

Contains utilities functions.

"""

import tomllib
import warnings

import numpy as np
import pandas as pd
from brainglobe_atlasapi import BrainGlobeAtlas


def get_starter_cells(animal: str, channel: str, info_file: str) -> int:
    """
    Get the number of starter cells associated with animal.

    Parameters
    ----------
    animal : str
        Animal ID.
    channel : str
        Channel ID.
    info_file : str
        Path to TOML info file.

    Returns
    -------
    n_starters : int
        Number of starter cells.

    """
    with open(info_file, "rb") as fid:
        info = tomllib.load(fid)

    return info[animal][channel]["starter_cells"]


def get_injection_site(
    animal: str, info_file: str, channel: str, stereo: bool = False
) -> tuple:
    """
    Get the injection site coordinates associated with animal.

    Parameters
    ----------
    animal : str
        Animal ID.
    info_file : str
        Path to TOML info file.
    channel : str
        Channel ID as in the TOML file.
    stereo : bool, optional
        Wether to convert coordinates in stereotaxis coordinates. Default is False.

    Returns
    -------
    x, y, z : floats
        Injection site coordinates.

    """
    with open(info_file, "rb") as fid:
        info = tomllib.load(fid)

    if channel in info[animal]:
        x, y, z = info[animal][channel]["injection_site"]
        if stereo:
            x, y, z = ccf_to_stereo(x, y, z)
    else:
        x, y, z = None, None, None

    return x, y, z


def get_mapping_fusion(fusion_file: str) -> dict:
    """
    Get mapping dictionnary between input brain regions and new regions defined in
    `atlas_fusion.toml` file.

    The returned dictionnary can be used in DataFrame.replace().

    Parameters
    ----------
    fusion_file : str
        Path to the TOML file with the merging rules.

    Returns
    -------
    m : dict
        Mapping as {old: new}.

    """
    with open(fusion_file, "rb") as fid:
        df = pd.DataFrame.from_dict(tomllib.load(fid), orient="index").set_index(
            "acronym"
        )

    return (
        df.drop(columns="name")["members"]
        .explode()
        .reset_index()
        .set_index("members")
        .to_dict()["acronym"]
    )


def get_blacklist(file: str, atlas: BrainGlobeAtlas | None) -> list:
    """
    Build a list of regions to exclude from file.

    File must be a TOML with [WITH_CHILDS] and [EXACT] sections. If `atlas` is None,
    return an empty list.

    Parameters
    ----------
    file : str
        Full path the atlas_blacklist.toml file.
    atlas : BrainGlobeAtlas or None
        Atlas to extract regions from.

    Returns
    -------
    black_list : list
        Full list of acronyms to discard.

    """
    # if no atlas provided, return empty list
    if atlas is None:
        return []

    with open(file, "rb") as fid:
        content = tomllib.load(fid)

    blacklist = []  # init. the list

    # add regions and their descendants
    for region in content["WITH_CHILDS"]["members"]:
        blacklist.extend(get_child_regions(atlas, region))

    # add regions specified exactly (no descendants)
    blacklist.extend(content["EXACT"]["members"])

    return blacklist


def merge_regions(df: pd.DataFrame, col: str, fusion_file: str) -> pd.DataFrame:
    """
    Merge brain regions following rules in the `fusion_file.toml` file.

    Apply this merging on `col` of the input DataFrame. `col` whose value is found in
    the `members` sections in the file will be changed to the new acronym.

    Parameters
    ----------
    df : pandas.DataFrame
    col : str
        Column of `df` on which to apply the mapping.
    fusion_file : str
        Path to the toml file with the merging rules.

    Returns
    -------
    df : pandas.DataFrame
        Same DataFrame with regions renamed.

    """
    df[col] = df[col].replace(get_mapping_fusion(fusion_file))

    return df


def get_leaves_list(atlas: BrainGlobeAtlas | None) -> list:
    """
    Get the list of leaf brain regions.

    Leaf brain regions are defined as regions without childs, eg. regions that are at
    the bottom of the hiearchy. If no atlas is provided, returns an empty list.

    Parameters
    ----------
    atlas : BrainGlobeAtlas or None
        Atlas to extract regions from.

    Returns
    -------
    leaves_list : list
        Acronyms of leaf brain regions.

    """
    leaves_list = []

    if atlas is None:
        return leaves_list

    for region in atlas.structures_list:
        if atlas.structures.tree[region["id"]].is_leaf():
            leaves_list.append(region["acronym"])

    return leaves_list


def get_child_regions(atlas: BrainGlobeAtlas | None, parent_region: str) -> list:
    """
    Get list of regions that are child of `parent_region`.

    If no atlas is provided, returns an empty list.

    Parameters
    ----------
    atlas : BrainGlobeAtlas or None
        Atlas to extract regions from.

    Returns
    -------
    child_list : list
        List of regions.

    """
    if atlas is None:
        return []

    return [
        atlas.structures[id]["acronym"]
        for id in atlas.structures.tree.expand_tree(
            atlas.structures[parent_region]["id"]
        )
    ]


def ccf_to_stereo(
    x_ccf: float | np.ndarray, y_ccf: float | np.ndarray, z_ccf: float | np.ndarray = 0
) -> tuple:
    """
    Convert X, Y, Z coordinates in CCFv3 to stereotaxis coordinates (as in
    Paxinos-Franklin atlas).

    Coordinates are shifted, rotated and squeezed, see (1) for more info. Input must be
    in mm.
    `x_ccf` corresponds to the anterio-posterior (rostro-caudal) axis.
    `y_ccf` corresponds to the dorso-ventral axis.
    `z_ccf` corresponds to the medio-lateral axis (left-right) axis.

    Warning : it is a rough estimation.

    (1) https://community.brain-map.org/t/how-to-transform-ccf-x-y-z-coordinates-into-stereotactic-coordinates/1858

    Parameters
    ----------
    x_ccf, y_ccf : floats or np.ndarray
        Coordinates in CCFv3 space in mm.
    z_ccf : float or np.ndarray, optional
        Coordinate in CCFv3 space in mm. Default is 0.

    Returns
    -------
    ap, dv, ml : floats or np.ndarray
        Stereotaxic coordinates in mm.

    """
    # Center CCF on Bregma
    xstereo = -(x_ccf - 5.40)  # anterio-posterior coordinate (rostro-caudal)
    ystereo = y_ccf - 0.44  # dorso-ventral coordinate
    ml = z_ccf - 5.70  # medio-lateral coordinate (left-right)

    # Rotate CCF of 5°
    angle = np.deg2rad(5)
    ap = xstereo * np.cos(angle) - ystereo * np.sin(angle)
    dv = xstereo * np.sin(angle) + ystereo * np.cos(angle)

    # Squeeze the dorso-ventral axis by 94.34%
    dv *= 0.9434

    return ap, dv, ml


def get_df_kind(df: pd.DataFrame) -> str:
    """
    Get DataFrame kind, eg. Annotations or Detections.

    It is based on reading the Object Type of the first entry, so the DataFrame must
    have only one kind of object.

    Parameters
    ----------
    df : pandas.DataFrame

    Returns
    -------
    kind : str
        "detection" or "annotation".

    """
    return df["Object type"].iloc[0].lower()


def add_hemisphere(
    df: pd.DataFrame,
    hemisphere_names: dict,
    midline: float = 5700,
    col: str = "Atlas_Z",
    atlas_type: str = "brain",
) -> pd.DataFrame:
    """
    Add hemisphere (left/right) as a measurement for detections or annotations.

    The hemisphere is read in the "Classification" column for annotations. The latter
    needs to be in the form "Right: Name" or "Left: Name". For detections, the input
    `col` of `df` is compared to `midline` to assess if the object belong to the left or
    right hemispheres.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame with detections or annotations measurements.
    hemisphere_names : dict
        Map between "Left" and "Right" to something else.
    midline : float
        Used only for "detections" `df`. Corresponds to the brain midline in microns,
        should be 5700 for CCFv3 and 1610 for spinal cord.
    col : str, optional
        Name of the column containing the Z coordinate (medio-lateral) in microns.
        Default is "Atlas_Z".
    atlas_type : {"brain", "cord"}, optional
        Type of atlas used for registration. Required because the brain atlas is swapped
        between left and right while the spinal cord atlas is not. Default is "brain".

    Returns
    -------
    df : pandas.DataFrame
        The same DataFrame with a new "hemisphere" column

    """
    # check if there is something to do
    if "hemisphere" in df.columns:
        return df

    # get kind of DataFrame
    kind = get_df_kind(df)

    if kind == "detection":
        # use midline
        if atlas_type == "brain":
            # brain atlas : beyond midline, it's left
            df.loc[df[col] >= midline, "hemisphere"] = hemisphere_names["Left"]
            df.loc[df[col] < midline, "hemisphere"] = hemisphere_names["Right"]
        elif atlas_type == "cord":
            # cord atlas : below midline, it's left
            df.loc[df[col] <= midline, "hemisphere"] = hemisphere_names["Left"]
            df.loc[df[col] > midline, "hemisphere"] = hemisphere_names["Right"]

    elif kind == "annotation":
        # use Classification name -- this does not depend on atlas type
        df["hemisphere"] = [name.split(":")[0] for name in df["Classification"]]
        df["hemisphere"] = df["hemisphere"].map(hemisphere_names)

    return df


def add_channel(
    df: pd.DataFrame, object_type: str, channel_names: dict
) -> pd.DataFrame:
    """
    Add channel as a measurement for detections DataFrame.

    The channel is read from the Classification column, the latter having to be
    formatted as "object_type: channel".

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with detections measurements.
    object_type : str
        Object type (primary classification).
    channel_names : dict
        Map between original channel names to something else.

    Returns
    -------
    pd.DataFrame
        Same DataFrame with a "channel" column.

    """
    # check if there is something to do
    if "channel" in df.columns:
        return df

    kind = get_df_kind(df)
    if kind == "annotation":
        warnings.warn("Annotation DataFrame not supported.")
        return df

    # add channel, from {class_name: channel} classification
    df["channel"] = (
        df["Classification"].str.replace(object_type + ": ", "").map(channel_names)
    )

    return df


def add_brain_region(
    df: pd.DataFrame, atlas: BrainGlobeAtlas | None, col="Parent"
) -> pd.DataFrame:
    """
    Add brain region to a DataFrame with `Atlas_X`, `Atlas_Y` and `Atlas_Z` columns.

    This uses Brainglobe Atlas API to query the atlas. It does not use the
    structure_from_coords() method, instead it manually converts the coordinates in
    stack indices, then get the corresponding annotation id and query the corresponding
    acronym -- because brainglobe-atlasapi is not vectorized at all.
    If no altas is provided (None), the `col` column is set to an empty string.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with atlas coordinates in microns.
    atlas : BrainGlobeAtlas or None
    col : str, optional
        Column in which to put the regions acronyms. Default is "Parent".

    Returns
    -------
    df : pd.DataFrame
        Same DataFrame with a new "Parent" column.

    """
    df_in = df.copy()

    if atlas is None:
        # no atlas provided set required col as empty string
        df[col] = ""
        return df

    res = atlas.resolution  # microns <-> pixels conversion
    lims = atlas.shape_um  # out of brain

    # set out-of-brain objects at 0 so we get "root" as their parent
    df_in.loc[(df_in["Atlas_X"] >= lims[0]) | (df_in["Atlas_X"] < 0), "Atlas_X"] = 0
    df_in.loc[(df_in["Atlas_Y"] >= lims[1]) | (df_in["Atlas_Y"] < 0), "Atlas_Y"] = 0
    df_in.loc[(df_in["Atlas_Z"] >= lims[2]) | (df_in["Atlas_Z"] < 0), "Atlas_Z"] = 0

    # build the multi index, in pixels and integers
    ixyz = (
        df_in["Atlas_X"].divide(res[0]).astype(int),
        df_in["Atlas_Y"].divide(res[1]).astype(int),
        df_in["Atlas_Z"].divide(res[2]).astype(int),
    )
    # convert i, j, k indices in raveled indices
    linear_indices = np.ravel_multi_index(ixyz, dims=atlas.annotation.shape)
    # get the structure id from the annotation stack
    idlist = atlas.annotation.ravel()[linear_indices]
    # replace 0 which does not exist to 997 (root)
    idlist[idlist == 0] = 997

    # query the corresponding acronyms
    lookup = atlas.lookup_df.set_index("id")
    df.loc[:, col] = lookup.loc[idlist, "acronym"].values

    return df


def filter_df_classifications(
    df: pd.DataFrame, filter_list: list | tuple | str, mode="keep", col="Classification"
) -> pd.DataFrame:
    """
    Filter a DataFrame whether specified `col` column entries contain elements in
    `filter_list`. Case insensitive.

    If `mode` is "keep", keep entries only if their `col` in is in the list (default).
    If `mode` is "remove", remove entries if their `col` is in the list.

    Parameters
    ----------
    df : pd.DataFrame
    filter_list : list | tuple | str
        List of words that should be present to trigger the filter.
    mode : "keep" or "remove", optional
        Keep or remove entries from the list. Default is "keep".
    col : str, optional
        Key in `df`. Default is "Classification".

    Returns
    -------
    pd.DataFrame
        Filtered DataFrame.

    """
    # check input
    if isinstance(filter_list, str):
        filter_list = [filter_list]  # make sure it is a list

    if col not in df.columns:
        # might be because of 'Classification' instead of 'classification'
        col = col.capitalize()
        if col not in df.columns:
            raise KeyError(f"{col} not in DataFrame.")

    pattern = "|".join(f".*{s}.*" for s in filter_list)

    if mode == "keep":
        df_return = df[df[col].str.contains(pattern, case=False, regex=True)]
    elif mode == "remove":
        df_return = df[~df[col].str.contains(pattern, case=False, regex=True)]

    # check
    if len(df_return) == 0:
        raise ValueError(
            (
                f"Filtering '{col}' with {filter_list} resulted in an"
                + " empty DataFrame, check your config file."
            )
        )
    return df_return


def filter_df_regions(
    df: pd.DataFrame, filter_list: list | tuple, mode="keep", col="Parent"
) -> pd.DataFrame:
    """
    Filters entries in `df` based on wether their `col` is in `filter_list` or not.

    If `mode` is "keep", keep entries only if their `col` in is in the list (default).
    If `mode` is "remove", remove entries if their `col` is in the list.

    Parameters
    ----------
    df : pandas.DataFrame
    filter_list : list-like
        List of regions to keep or remove from the DataFrame.
    mode : "keep" or "remove", optional
        Keep or remove entries from the list. Default is "keep".
    col : str, optional
        Key in `df`. Default is "Parent".

    Returns
    -------
    df : pandas.DataFrame
        Filtered DataFrame.

    """

    if mode == "keep":
        return df[df[col].isin(filter_list)]
    if mode == "remove":
        return df[~df[col].isin(filter_list)]


def get_data_coverage(df: pd.DataFrame, col="Atlas_AP", by="animal") -> pd.DataFrame:
    """
    Get min and max in `col` for each `by`.

    Used to get data coverage for each animal to plot in distributions.

    Parameters
    ----------
    df : pd.DataFrame
        _description_
    col : str, optional
        Key in `df`, default is "Atlas_X".
    by : str, optional
        Key in `df` , default is "animal".

    Returns
    -------
    pd.DataFrame
        min and max of `col` for each `by`, named "X_min", and "X_max".

    """
    df_group = df.groupby([by])
    return pd.DataFrame(
        [
            df_group[col].min(),
            df_group[col].max(),
        ],
        index=["X_min", "X_max"],
    )


def renormalize_per_key(df: pd.DataFrame, by: str, on: str):
    """
    Renormalize `on` column by its sum for each `by`.

    Use case : relative density is computed for both hemispheres, so if one wants to
    plot only one hemisphere, the sum of the bars corresponding to one channel (`by`)
    should be 1. So :
    >>> df = df[df["hemisphere"] == "Ipsi."]
    >>> df = renormalize_per_key(df, "channel", "relative density")
    Then, the sum of "relative density" for each "channel" equals 1.

    Parameters
    ----------
    df : pd.DataFrame
    by : str
        Key in `df`. `df` is normalized for each `by`.
    on : str
        Key in `df`. Measurement to be normalized.

    Returns
    -------
    df : pd.DataFrame
        Same DataFrame with normalized `on` column.

    """
    norm = df.groupby(by)[on].sum()
    bys = df[by].unique()
    for key in bys:
        df.loc[df[by] == key, on] = df.loc[df[by] == key, on].divide(norm[key])

    return df


def select_hemisphere_channel(
    df: pd.DataFrame, hue: str, hue_filter: str, hue_mirror: bool
) -> pd.DataFrame:
    """
    Select relevant data given hue and filters.

    Returns the DataFrame with only things to be used.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to filter.
    hue : {"hemisphere", "channel"}
        hue that will be used in seaborn plots.
    hue_filter : str
        Selected data.
    hue_mirror : bool
        Instead of keeping only hue_filter values, they will be plotted in mirror.

    Returns
    -------
    dfplt : pd.DataFrame
        DataFrame to be used in plots.

    """
    dfplt = df.copy()

    if hue == "hemisphere":
        # hue_filter is used to select channels
        # keep only left and right hemispheres, not "both"
        dfplt = dfplt[dfplt["hemisphere"] != "both"]
        if hue_filter == "all":
            hue_filter = dfplt["channel"].unique()
        elif not isinstance(hue_filter, (list, tuple)):
            # it is allowed to select several channels so handle lists
            hue_filter = [hue_filter]
        dfplt = dfplt[dfplt["channel"].isin(hue_filter)]
    elif hue == "channel":
        # hue_filter is used to select hemispheres
        # it can only be left, right, both or empty
        if hue_filter == "both":
            # handle if it's a coordinates DataFrame which doesn't have "both"
            if "both" not in dfplt["hemisphere"].unique():
                # keep both hemispheres, don't do anything
                pass
            else:
                if hue_mirror:
                    # we need to keep both hemispheres to plot them in mirror
                    dfplt = dfplt[dfplt["hemisphere"] != "both"]
                else:
                    # we keep the metrics computed in both hemispheres
                    dfplt = dfplt[dfplt["hemisphere"] == "both"]
        else:
            # hue_filter should correspond to an hemisphere name
            dfplt = dfplt[dfplt["hemisphere"] == hue_filter]
    else:
        # not handled. Just return the DataFrame without filtering, maybe it'll make
        # sense.
        warnings.warn(f"{hue} should be 'channel' or 'hemisphere'.")

    # check result
    if len(dfplt) == 0:
        warnings.warn(
            f"hue={hue} and hue_filter={hue_filter} resulted in an empty subset."
        )

    return dfplt



"""process module, part of cuisto.

Wraps other functions for a click&play behaviour. Relies on the configuration file.

"""

import os

import pandas as pd
from tqdm import tqdm

from cuisto import compute, io, utils


def process_animal(
    animal: str,
    df_annotations: pd.DataFrame,
    df_detections: pd.DataFrame,
    cfg,
    compute_distributions: bool = True,
    leaf_regions_only: bool = True,
) -> tuple[pd.DataFrame, list[pd.DataFrame], pd.DataFrame]:
    """
    Quantify objects for one animal.

    Fetch required files and compute objects' distributions in brain regions, spatial
    distributions and gather Atlas coordinates.

    Parameters
    ----------
    animal : str
        Animal ID.
    df_annotations, df_detections : pd.DataFrame
        DataFrames of QuPath Annotations and Detections.
    cfg : cuisto.Config
        The configuration loaded from TOML configuration file.
    compute_distributions : bool, optional
        If False, do not compute the 1D distributions and return an empty list. Default
        is True.
    leaf_regions_only : bool, optional
        If True and a Brainglobe atlas is specified, bar plot per regions will keep only
        leaf regions, eg. regions with no child -- if there are any. Default is True.

    Returns
    -------
    df_regions : pandas.DataFrame
        Metrics in brain regions. One entry for each hemisphere of each brain regions.
    df_distribution : list of pandas.DataFrame
        Rostro-caudal distribution, as raw count and probability density function, in
        each axis.
    df_coordinates : pandas.DataFrame
        Atlas coordinates of each points.

    """
    # - Annotations data cleanup
    # filter regions
    df_annotations = utils.filter_df_regions(
        df_annotations, ["Root", "root"], mode="remove", col="Name"
    )
    df_annotations = utils.filter_df_regions(
        df_annotations, cfg.atlas["blacklist"], mode="remove", col="Name"
    )
    # add hemisphere
    df_annotations = utils.add_hemisphere(df_annotations, cfg.hemispheres["names"])
    # remove objects in non-leaf regions if any
    if leaf_regions_only & (len(cfg.atlas["leaveslist"]) > 0):
        df_annotations = utils.filter_df_regions(
            df_annotations, cfg.atlas["leaveslist"], mode="keep", col="Name"
        )
    # merge regions
    df_annotations = utils.merge_regions(
        df_annotations, col="Name", fusion_file=cfg.files["fusion"]
    )
    if compute_distributions:
        # - Detections data cleanup
        # remove objects not in selected classifications
        df_detections = utils.filter_df_classifications(
            df_detections, cfg.object_type, mode="keep", col="Classification"
        )
        # remove objects from blacklisted regions
        df_detections = utils.filter_df_regions(
            df_detections, cfg.atlas["blacklist"], mode="remove", col="Parent"
        )
        # add hemisphere
        df_detections = utils.add_hemisphere(
            df_detections,
            cfg.hemispheres["names"],
            cfg.atlas["midline"],
            col="Atlas_Z",
            atlas_type=cfg.atlas["type"],
        )
        # add detection channel
        df_detections = utils.add_channel(
            df_detections, cfg.object_type, cfg.channels["names"]
        )
        # convert coordinates to mm
        df_detections[["Atlas_X", "Atlas_Y", "Atlas_Z"]] = df_detections[
            ["Atlas_X", "Atlas_Y", "Atlas_Z"]
        ].divide(1000)
        # convert to sterotaxic coordinates
        if cfg.distributions["stereo"]:
            (
                df_detections["Atlas_AP"],
                df_detections["Atlas_DV"],
                df_detections["Atlas_ML"],
            ) = utils.ccf_to_stereo(
                df_detections["Atlas_X"],
                df_detections["Atlas_Y"],
                df_detections["Atlas_Z"],
            )
        else:
            (
                df_detections["Atlas_AP"],
                df_detections["Atlas_DV"],
                df_detections["Atlas_ML"],
            ) = (
                df_detections["Atlas_X"],
                df_detections["Atlas_Y"],
                df_detections["Atlas_Z"],
            )

    # - Computations
    # get regions distributions
    df_regions = compute.get_regions_metrics(
        df_annotations,
        cfg.object_type,
        cfg.channels["names"],
        cfg.regions["base_measurement"],
        cfg.regions["metrics"],
    )
    colstonorm = [v for v in cfg.regions["metrics"].values() if "relative" not in v]

    # normalize by starter cells
    if cfg.regions["normalize_starter_cells"]:
        df_regions = compute.normalize_starter_cells(
            df_regions, colstonorm, animal, cfg.files["infos"], cfg.channels["names"]
        )

    # get AP, DV, ML distributions in stereotaxic coordinates
    if compute_distributions:
        dfs_distributions = [
            compute.get_distribution(
                df_detections,
                axis,
                cfg.distributions["hue"],
                cfg.distributions["hue_filter"],
                cfg.distributions["common_norm"],
                stereo_lim,
                nbins=nbins,
            )
            for axis, stereo_lim, nbins in zip(
                ["Atlas_AP", "Atlas_DV", "Atlas_ML"],
                [
                    cfg.distributions["ap_lim"],
                    cfg.distributions["dv_lim"],
                    cfg.distributions["ml_lim"],
                ],
                [
                    cfg.distributions["ap_nbins"],
                    cfg.distributions["dv_nbins"],
                    cfg.distributions["dv_nbins"],
                ],
            )
        ]
    else:
        dfs_distributions = []

    # add animal tag to each DataFrame
    df_detections["animal"] = animal
    df_regions["animal"] = animal
    for df in dfs_distributions:
        df["animal"] = animal

    return df_regions, dfs_distributions, df_detections


def process_animals(
    wdir: str,
    animals: list[str] | tuple[str],
    cfg,
    out_fmt: str | None = None,
    compute_distributions: bool = True,
    **kwargs,
) -> tuple[pd.DataFrame]:
    """
    Get data from all animals and plot.

    Parameters
    ----------
    wdir : str
        Base working directory, containing `animals` folders.
    animals : list-like of str
        List of animals ID.
    cfg: cuisto.Config
        Configuration object.
    out_fmt : {None, "h5", "csv", "tsv", "xslx", "pickle"}
        Output file(s) format, if None, nothing is saved (default).
    compute_distributions : bool, optional
        If False, do not compute the 1D distributions and return an empty list.Default
        is True.
    kwargs : passed to cuisto.process.process_animal().

    Returns
    -------
    df_regions : pandas.DataFrame
        Metrics in brain regions. One entry for each hemisphere of each brain regions.
    df_distribution : list of pandas.DataFrame
        Rostro-caudal distribution, as raw count and probability density function, in
        each axis.
    df_coordinates : pandas.DataFrame
        Atlas coordinates of each points.

    """

    # -- Preparation
    df_regions = []
    dfs_distributions = []
    df_coordinates = []

    # -- Processing
    pbar = tqdm(animals)

    for animal in pbar:
        pbar.set_description(f"Processing {animal}")

        # combine all detections and annotations from this animal
        df_annotations = io.cat_csv_dir(
            io.get_measurements_directory(
                wdir, animal, "annotation", cfg.segmentation_tag
            ),
            index_col="Object ID",
            sep="\t",
        )
        if compute_distributions:
            df_detections = io.cat_data_dir(
                io.get_measurements_directory(
                    wdir, animal, "detection", cfg.segmentation_tag
                ),
                cfg.segmentation_tag,
                index_col="Object ID",
                sep="\t",
                hemisphere_names=cfg.hemispheres["names"],
                atlas=cfg.bg_atlas,
            )
        else:
            df_detections = pd.DataFrame()

        # get results
        df_reg, dfs_dis, df_coo = process_animal(
            animal,
            df_annotations,
            df_detections,
            cfg,
            compute_distributions=compute_distributions,
            **kwargs,
        )

        # collect results
        df_regions.append(df_reg)
        dfs_distributions.append(dfs_dis)
        df_coordinates.append(df_coo)

    # concatenate all results
    df_regions = pd.concat(df_regions, ignore_index=True)
    dfs_distributions = [
        pd.concat(dfs_list, ignore_index=True) for dfs_list in zip(*dfs_distributions)
    ]
    df_coordinates = pd.concat(df_coordinates, ignore_index=True)

    # -- Saving
    if out_fmt:
        outdir = os.path.join(wdir, "quantification")
        outfile = f"{cfg.object_type.lower()}_{cfg.atlas["type"]}_{'-'.join(animals)}.{out_fmt}"
        dfs = dict(
            df_regions=df_regions,
            df_coordinates=df_coordinates,
            df_distribution_ap=dfs_distributions[0],
            df_distribution_dv=dfs_distributions[1],
            df_distribution_ml=dfs_distributions[2],
        )
        io.save_dfs(outdir, outfile, dfs)

    return df_regions, dfs_distributions, df_coordinates



"""display module, part of cuisto.

Contains display functions, essentially wrapping matplotlib and seaborn functions.

"""

import warnings

import h5py
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib import patches

from cuisto import utils


def add_injection_patch(X: list, ax: plt.Axes, **kwargs) -> plt.Axes:
    """
    Add a patch representing the injection sites.

    The patch will span from the minimal coordinate to the maximal.
    If plotted in stereotaxic coordinates, coordinates should be converted beforehand.

    Parameters
    ----------
    X : list
        Coordinates in mm for each animals. Can be empty to not plot anything.
    ax : Axes
        Handle to axes where to add the patch.
    **kwargs : passed to Axes.axvspan

    Returns
    -------
    ax : Axes
        Handle to updated Axes.

    """
    # plot patch
    if len(X) > 0:
        ax.axvspan(min(X), max(X), **kwargs)

    return ax


def add_data_coverage(
    df: pd.DataFrame, ax: plt.Axes, colors: list | str | None = None, **kwargs
) -> plt.Axes:
    """
    Add lines below the plot to represent data coverage.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame with `X_min` and `X_max` on rows for each animals (on columns).
    ax : Axes
        Handle to axes where to add the patch.
    colors : list or str or None, optional
        Colors for the patches, as a RGB list or hex list. Should be the same size as
        the number of patches to plot, eg. the number of columns in `df`. If None,
        default seaborn colors are used. If only one element, used for each animal.
    **kwargs : passed to patches.Rectangle()

    Returns
    -------
    ax : Axes
        Handle to updated axes.

    """
    # get colors
    ncolumns = len(df.columns)
    if not colors:
        colors = sns.color_palette(n_colors=ncolumns)
    elif isinstance(colors, str) or (isinstance(colors, list) & (len(colors) == 3)):
        colors = [colors] * ncolumns
    elif len(colors) != ncolumns:
        warnings.warn(f"Wrong number of colors ({len(colors)}), using default colors.")
        colors = sns.color_palette(n_colors=ncolumns)

    # get patch height depending on current axis limits
    ymin, ymax = ax.get_ylim()
    height = (ymax - ymin) * 0.02

    for animal, color in zip(df.columns, colors):
        # get patch coordinates
        ymin, ymax = ax.get_ylim()
        ylength = ymax - ymin
        ybottom = ymin - 0.02 * ylength
        xleft = df.loc["X_min", animal]
        xright = df.loc["X_max", animal]

        # plot patch
        ax.add_patch(
            patches.Rectangle(
                (xleft, ybottom),
                xright - xleft,
                height,
                label=animal,
                color=color,
                **kwargs,
            )
        )

        ax.autoscale(tight=True)  # set new axes limits

    ax.autoscale()  # reset scale

    return ax


def draw_structure_outline(
    view: str = "sagittal",
    structures: list[str] = ["root"],
    outline_file: str = "",
    ax: plt.Axes | None = None,
    microns: bool = False,
    **kwargs,
) -> plt.Axes:
    """
    Plot brain regions outlines in given projection.

    This requires a file containing the structures outlines.

    Parameters
    ----------
    view : str
        Projection, "sagittal", "coronal" or "top". Default is "sagittal".
    structures : list[str]
        List of structures acronyms whose outlines will be drawn. Default is ["root"].
    outline_file : str
        Full path the outlines HDF5 file.
    ax : plt.Axes or None, optional
        Axes where to plot the outlines. If None, get current axes (the default).
    microns : bool, optional
        If False (default), converts the coordinates in mm.
    **kwargs : passed to pyplot.plot()

    Returns
    -------
    ax : plt.Axes

    """
    # get axes
    if not ax:
        ax = plt.gca()

    # get units
    if microns:
        conv = 1
    else:
        conv = 1 / 1000

    with h5py.File(outline_file) as f:
        if view == "sagittal":
            for structure in structures:
                dsets = f["sagittal"][structure]

                for dset in dsets.values():
                    ax.plot(dset[:, 0] * conv, dset[:, 1] * conv, **kwargs)

        if view == "coronal":
            for structure in structures:
                dsets = f["coronal"][structure]

                for dset in dsets.values():
                    ax.plot(dset[:, 0] * conv, dset[:, 1] * conv, **kwargs)

        if view == "top":
            for structure in structures:
                dsets = f["top"][structure]

                for dset in dsets.values():
                    ax.plot(dset[:, 0] * conv, dset[:, 1] * conv, **kwargs)

    return ax


def nice_bar_plot(
    df: pd.DataFrame,
    x: str = "",
    y: list[str] = [""],
    hue: str = "",
    ylabel: list[str] = [""],
    orient="h",
    nx: None | int = None,
    ordering: None | list[str] | str = None,
    names_list: None | list = None,
    hue_mirror: bool = False,
    log_scale: bool = False,
    bar_kws: dict = {},
    pts_kws: dict = {},
) -> list[plt.Axes]:
    """
    Nice bar plot of per-region objects distribution.

    This is used for objects distribution across brain regions. Shows the `y` metric
    (count, aeral density, cumulated length...) in each `x` categories (brain regions).
    `orient` controls wether the bars are shown horizontally (default) or vertically.
    Input `df` must have an additional "hemisphere" column. All `y` are plotted in the
    same figure as different subplots. `nx` controls the number of displayed regions.

    Parameters
    ----------
    df : pandas.DataFrame
    x, y, hue : str
        Key in `df`.
    ylabel : list of str
        Y axis labels.
    orient : "h" or "v", optional
        "h" for horizontal bars (default) or "v" for vertical bars.
    nx : None or int, optional
        Number of `x` to show in the plot. Default is None (no limit).
    ordering : None or list[str] or "max", optional
        Sorted list of acronyms. Data will be sorted follwowing this order, if "max",
        sorted by descending values, if None, not sorted (default).
    names_list : list or None, optional
        List of names to display. If None (default), takes the most prominent overall
        ones.
    hue_mirror : bool, optional
        If there are 2 groups, plot in mirror. Default is False.
    log_scale : bool, optional
        Set the metrics in log scale. Default is False.
    bar_kws : dict
        Passed to seaborn.barplot().
    pts_kws : dict
        Passed to seaborn.stripplot().

    Returns
    -------
    figs : list
        List of figures.

    """
    figs = []
    # loop for each features
    for yi, ylabeli in zip(y, ylabel):
        # prepare data
        # get nx first most prominent regions
        if not names_list:
            names_list_plt = (
                df.groupby(["Name"])[yi].mean().sort_values(ascending=False).index[0:nx]
            )
        else:
            names_list_plt = names_list
        dfplt = df[df["Name"].isin(names_list_plt)]  # limit to those regions
        # limit hierarchy list if provided
        if isinstance(ordering, list):
            order = [el for el in ordering if el in names_list_plt]
        elif ordering == "max":
            order = names_list_plt
        else:
            order = None

        # reorder keys depending on orientation and create axes
        if orient == "h":
            xp = yi
            yp = x
            if hue_mirror:
                nrows = 1
                ncols = 2
                sharex = None
                sharey = "all"
            else:
                nrows = 1
                ncols = 1
                sharex = None
                sharey = None
        elif orient == "v":
            xp = x
            yp = yi
            if hue_mirror:
                nrows = 2
                ncols = 1
                sharex = "all"
                sharey = None
            else:
                nrows = 1
                ncols = 1
                sharex = None
                sharey = None
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, sharex=sharex, sharey=sharey)

        if hue_mirror:
            # two graphs
            ax1, ax2 = axs
            # determine what will be mirrored
            if hue == "channel":
                hue_filter = "hemisphere"
            elif hue == "hemisphere":
                hue_filter = "channel"
            # select the two types (should be left/right or two channels)
            hue_filters = dfplt[hue_filter].unique()[0:2]
            hue_filters.sort()  # make sure it will be always in the same order

            # plot
            for filt, ax in zip(hue_filters, [ax1, ax2]):
                dfplt2 = dfplt[dfplt[hue_filter] == filt]
                ax = sns.barplot(
                    dfplt2,
                    x=xp,
                    y=yp,
                    hue=hue,
                    estimator="mean",
                    errorbar="se",
                    orient=orient,
                    order=order,
                    ax=ax,
                    **bar_kws,
                )
                # add points
                ax = sns.stripplot(
                    dfplt2, x=xp, y=yp, hue=hue, legend=False, ax=ax, **pts_kws
                )

                # cosmetics
                if orient == "h":
                    ax.set_title(f"{hue_filter}: {filt}")
                    ax.set_ylabel(None)
                    ax.set_ylim((nx + 0.5, -0.5))
                    if log_scale:
                        ax.set_xscale("log")

                elif orient == "v":
                    if ax == ax1:
                        # top title
                        ax1.set_title(f"{hue_filter}: {filt}")
                        ax.set_xlabel(None)
                    elif ax == ax2:
                        # use xlabel as bottom title
                        ax2.set_xlabel(
                            f"{hue_filter}: {filt}", fontsize=ax1.title.get_fontsize()
                        )
                    ax.set_xlim((-0.5, nx + 0.5))
                    if log_scale:
                        ax.set_yscale("log")

                    for label in ax.get_xticklabels():
                        label.set_verticalalignment("center")
                        label.set_horizontalalignment("center")

            # tune axes cosmetics
            if orient == "h":
                ax1.set_xlabel(ylabeli)
                ax2.set_xlabel(ylabeli)
                ax1.set_xlim(
                    ax1.get_xlim()[0], max((ax1.get_xlim()[1], ax2.get_xlim()[1]))
                )
                ax2.set_xlim(
                    ax2.get_xlim()[0], max((ax1.get_xlim()[1], ax2.get_xlim()[1]))
                )
                ax1.invert_xaxis()
                sns.despine(ax=ax1, left=True, top=True, right=False, bottom=False)
                sns.despine(ax=ax2, left=False, top=True, right=True, bottom=False)
                ax1.yaxis.tick_right()
                ax1.tick_params(axis="y", pad=20)
                for label in ax1.get_yticklabels():
                    label.set_verticalalignment("center")
                    label.set_horizontalalignment("center")
            elif orient == "v":
                ax2.set_ylabel(ylabeli)
                ax1.set_ylim(
                    ax1.get_ylim()[0], max((ax1.get_ylim()[1], ax2.get_ylim()[1]))
                )
                ax2.set_ylim(
                    ax2.get_ylim()[0], max((ax1.get_ylim()[1], ax2.get_ylim()[1]))
                )
                ax2.invert_yaxis()
                sns.despine(ax=ax1, left=False, top=True, right=True, bottom=False)
                sns.despine(ax=ax2, left=False, top=False, right=True, bottom=True)
                for label in ax2.get_xticklabels():
                    label.set_verticalalignment("center")
                    label.set_horizontalalignment("center")
                ax2.tick_params(axis="x", labelrotation=90, pad=20)

        else:
            # one graph
            ax = axs
            # plot
            ax = sns.barplot(
                dfplt,
                x=xp,
                y=yp,
                hue=hue,
                estimator="mean",
                errorbar="se",
                orient=orient,
                order=order,
                ax=ax,
                **bar_kws,
            )
            # add points
            ax = sns.stripplot(
                dfplt, x=xp, y=yp, hue=hue, legend=False, ax=ax, **pts_kws
            )

            # cosmetics
            if orient == "h":
                ax.set_xlabel(ylabeli)
                ax.set_ylabel(None)
                ax.set_ylim((nx + 0.5, -0.5))
                if log_scale:
                    ax.set_xscale("log")
            elif orient == "v":
                ax.set_xlabel(None)
                ax.set_ylabel(ylabeli)
                ax.set_xlim((-0.5, nx + 0.5))
                if log_scale:
                    ax.set_yscale("log")

        fig.tight_layout(pad=0)
        figs.append(fig)

    return figs


def nice_distribution_plot(
    df: pd.DataFrame,
    x: str = "",
    y: str = "",
    hue: str | None = None,
    xlabel: str = "",
    ylabel: str = "",
    injections_sites: dict = {},
    channel_colors: dict = {},
    channel_names: dict = {},
    ax: plt.Axes | None = None,
    **kwargs,
) -> plt.Axes:
    """
    Nice plot of 1D distribution of objects.

    Parameters
    ----------
    df : pandas.DataFrame
    x, y : str
        Keys in `df`.
    hue : str or None, optional
        Key in `df`. If None, no hue is used.
    xlabel, ylabel : str
        X and Y axes labels.
    injections_sites : dict, optional
        List of injection sites 1D coordinates in a dict with the channel name as key.
        If empty, injection site is not plotted (default).
    channel_colors : dict, optional
        Required if injections_sites is not empty, dict mapping channel names to a
        color.
    channel_names : dict, optional
        Required if injections_sites is not empty, dict mapping channel names to a
        display name.
    ax : Axes or None, optional
        Axes in which to plot the figure, if None, a new figure is created (default).
    **kwargs : passed to seaborn.lineplot()

    Returns
    -------
    ax : matplotlib axes
        Handle to axes.

    """
    if not ax:
        # create figure
        _, ax = plt.subplots(figsize=(10, 6))

    ax = sns.lineplot(
        df,
        x=x,
        y=y,
        hue=hue,
        estimator="mean",
        errorbar="se",
        ax=ax,
        **kwargs,
    )

    for channel in injections_sites.keys():
        ax = add_injection_patch(
            injections_sites[channel],
            ax,
            color=channel_colors[channel],
            edgecolor=None,
            alpha=0.25,
            label=channel_names[channel] + ": inj. site",
        )

    ax.legend()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ax


def nice_joint_plot(
    df: pd.DataFrame,
    x: str = "",
    y: str = "",
    xlabel: str = "",
    ylabel: str = "",
    invertx: bool = False,
    inverty: bool = False,
    outline_kws: dict = {},
    ax: plt.Axes | None = None,
    **kwargs,
) -> plt.Figure:
    """
    Joint distribution.

    Used to display a 2D heatmap of objects. This is more qualitative than quantitative,
    for display purposes.

    Parameters
    ----------
    df : pandas.DataFrame
    x, y : str
        Keys in `df`.
    xlabel, ylabel : str
        Label of x and y axes.
    invertx, inverty : bool, optional
        Whether to inverse the x or y axes. Default is False for both.
    outline_kws : dict
        Passed to draw_structure_outline().
    ax : plt.Axes or None, optional
        Axes to plot in. If None, draws in current axes (default).
    **kwargs
        Passed to seaborn.histplot.

    Returns
    -------
    ax : plt.Axes

    """
    if not ax:
        ax = plt.gca()

    # plot outline
    draw_structure_outline(ax=ax, **outline_kws)

    # plot joint distribution
    sns.histplot(
        df,
        x=x,
        y=y,
        ax=ax,
        **kwargs,
    )

    # adjust axes
    if invertx:
        ax.invert_xaxis()
    if inverty:
        ax.invert_yaxis()

    # labels
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ax


def nice_heatmap(
    df: pd.DataFrame,
    animals: tuple[str] | list[str],
    x: str = "",
    y: str = "",
    xlabel: str = "",
    ylabel: str = "",
    invertx: bool = False,
    inverty: bool = False,
    **kwargs,
) -> list[plt.Axes] | plt.Axes:
    """
    Nice plots of 2D distribution of boutons as a heatmap per animal.

    Parameters
    ----------
    df : pandas.DataFrame
    animals : list-like of str
        List of animals.
    x, y : str
        Keys in `df`.
    xlabel, ylabel : str
        Labels of x and y axes.
    invertx, inverty : bool, optional
        Wether to inverse the x or y axes. Default is False.
    **kwargs : passed to seaborn.histplot()

    Returns
    -------
    ax : Axes or list of Axes
        Handle to axes.

    """

    # 2D distribution, per animal
    _, axs = plt.subplots(len(animals), 1, sharex="all")

    for animal, ax in zip(animals, axs):
        ax = sns.histplot(
            df[df["animal"] == animal],
            x=x,
            y=y,
            ax=ax,
            **kwargs,
        )
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(animal)

        if inverty:
            ax.invert_yaxis()

    if invertx:
        axs[-1].invert_xaxis()  # only once since all x axes are shared

    return axs


def plot_regions(df: pd.DataFrame, cfg, **kwargs):
    """
    Wraps nice_bar_plot().
    """
    # get regions order
    if cfg.regions["display"]["order"] == "ontology":
        regions_order = [d["acronym"] for d in cfg.bg_atlas.structures_list]
    elif cfg.regions["display"]["order"] == "max":
        regions_order = "max"
    else:
        regions_order = None

    # determine metrics to be plotted and color palette based on hue
    metrics = [*cfg.regions["display"]["metrics"].keys()]
    hue = cfg.regions["hue"]
    palette = cfg.get_hue_palette("regions")

    # select data
    dfplt = utils.select_hemisphere_channel(
        df, hue, cfg.regions["hue_filter"], cfg.regions["hue_mirror"]
    )

    # prepare options
    bar_kws = dict(
        err_kws={"linewidth": 1.5},
        dodge=cfg.regions["display"]["dodge"],
        palette=palette,
    )
    pts_kws = dict(
        size=4,
        edgecolor="auto",
        linewidth=0.75,
        dodge=cfg.regions["display"]["dodge"],
        palette=palette,
    )
    # draw
    figs = nice_bar_plot(
        dfplt,
        x="Name",
        y=metrics,
        hue=hue,
        ylabel=[*cfg.regions["display"]["metrics"].values()],
        orient=cfg.regions["display"]["orientation"],
        nx=cfg.regions["display"]["nregions"],
        ordering=regions_order,
        hue_mirror=cfg.regions["hue_mirror"],
        log_scale=cfg.regions["display"]["log_scale"],
        bar_kws=bar_kws,
        pts_kws=pts_kws,
        **kwargs,
    )

    return figs


def plot_1D_distributions(
    dfs_distributions: list[pd.DataFrame],
    cfg,
    df_coordinates: pd.DataFrame = None,
):
    """
    Wraps nice_distribution_plot().
    """
    # prepare figures
    fig, axs_dist = plt.subplots(1, 3, sharey=True, figsize=(13, 6))
    xlabels = [
        "Rostro-caudal position (mm)",
        "Dorso-ventral position (mm)",
        "Medio-lateral position (mm)",
    ]

    # get animals
    animals = []
    for df in dfs_distributions:
        animals.extend(df["animal"].unique())
    animals = set(animals)

    # get injection sites
    if cfg.distributions["display"]["show_injection"]:
        injection_sites = cfg.get_injection_sites(animals)
    else:
        injection_sites = {k: {} for k in range(3)}

    # get color palette based on hue
    hue = cfg.distributions["hue"]
    palette = cfg.get_hue_palette("distributions")

    # loop through each axis
    for df_dist, ax_dist, xlabel, inj_sites in zip(
        dfs_distributions, axs_dist, xlabels, injection_sites.values()
    ):
        # select data
        if cfg.distributions["hue"] == "hemisphere":
            dfplt = df_dist[df_dist["hemisphere"] != "both"]
        elif cfg.distributions["hue"] == "channel":
            dfplt = df_dist[df_dist["channel"] != "all"]

        # plot
        ax_dist = nice_distribution_plot(
            dfplt,
            x="bins",
            y="distribution",
            hue=hue,
            xlabel=xlabel,
            ylabel="normalized distribution",
            injections_sites=inj_sites,
            channel_colors=cfg.channels["colors"],
            channel_names=cfg.channels["names"],
            linewidth=2,
            palette=palette,
            ax=ax_dist,
        )

        # add data coverage
        if ("Atlas_AP" in df_dist["axis"].unique()) & (df_coordinates is not None):
            df_coverage = utils.get_data_coverage(df_coordinates)
            ax_dist = add_data_coverage(df_coverage, ax_dist, edgecolor=None, alpha=0.5)
            ax_dist.legend()
        else:
            ax_dist.legend().remove()

    # - Distributions, per animal
    if len(animals) > 1:
        _, axs_dist = plt.subplots(1, 3, sharey=True)

        # loop through each axis
        for df_dist, ax_dist, xlabel, inj_sites in zip(
            dfs_distributions, axs_dist, xlabels, injection_sites.values()
        ):
            # select data
            df_dist_plot = df_dist[df_dist["hemisphere"] == "both"]

            # plot
            ax_dist = nice_distribution_plot(
                df_dist_plot,
                x="bins",
                y="distribution",
                hue="animal",
                xlabel=xlabel,
                ylabel="normalized distribution",
                injections_sites=inj_sites,
                channel_colors=cfg.channels["colors"],
                channel_names=cfg.channels["names"],
                linewidth=2,
                ax=ax_dist,
            )

    return fig


def plot_2D_distributions(df: pd.DataFrame, cfg):
    """
    Wraps nice_joint_plot().
    """
    # -- 2D heatmap, all animals pooled
    # prepare figure
    fig_heatmap = plt.figure(figsize=(12, 9))

    ax_sag = fig_heatmap.add_subplot(2, 2, 1)
    ax_cor = fig_heatmap.add_subplot(2, 2, 2, sharey=ax_sag)
    ax_top = fig_heatmap.add_subplot(2, 2, 3, sharex=ax_sag)
    ax_cbar = fig_heatmap.add_subplot(2, 2, 4, box_aspect=15)

    # prepare options
    map_options = dict(
        bins=cfg.distributions["display"]["cmap_nbins"],
        cmap=cfg.distributions["display"]["cmap"],
        rasterized=True,
        thresh=10,
        stat="count",
        vmin=cfg.distributions["display"]["cmap_lim"][0],
        vmax=cfg.distributions["display"]["cmap_lim"][1],
    )
    outline_kws = dict(
        structures=cfg.atlas["outline_structures"],
        outline_file=cfg.files["outlines"],
        linewidth=1.5,
        color="k",
    )
    cbar_kws = dict(label="count")

    # determine which axes are going to be inverted
    if cfg.atlas["type"] == "brain":
        cor_invertx = True
        cor_inverty = False
        top_invertx = True
        top_inverty = False
    elif cfg.atlas["type"] == "cord":
        cor_invertx = False
        cor_inverty = False
        top_invertx = True
        top_inverty = True

    # - sagittal
    # no need to invert axes because they are shared with the two other views
    outline_kws["view"] = "sagittal"
    nice_joint_plot(
        df,
        x="Atlas_X",
        y="Atlas_Y",
        xlabel="Rostro-caudal (mm)",
        ylabel="Dorso-ventral (mm)",
        outline_kws=outline_kws,
        ax=ax_sag,
        **map_options,
    )

    # - coronal
    outline_kws["view"] = "coronal"
    nice_joint_plot(
        df,
        x="Atlas_Z",
        y="Atlas_Y",
        xlabel="Medio-lateral (mm)",
        ylabel="Dorso-ventral (mm)",
        invertx=cor_invertx,
        inverty=cor_inverty,
        outline_kws=outline_kws,
        ax=ax_cor,
        **map_options,
    )
    ax_cor.invert_yaxis()

    # - top
    outline_kws["view"] = "top"
    nice_joint_plot(
        df,
        x="Atlas_X",
        y="Atlas_Z",
        xlabel="Rostro-caudal (mm)",
        ylabel="Medio-lateral (mm)",
        invertx=top_invertx,
        inverty=top_inverty,
        outline_kws=outline_kws,
        ax=ax_top,
        cbar=True,
        cbar_ax=ax_cbar,
        cbar_kws=cbar_kws,
        **map_options,
    )
    fig_heatmap.suptitle("sagittal, coronal and top-view projections")

    # -- 2D heatmap per animals
    # get animals
    animals = df["animal"].unique()
    if len(animals) > 1:
        # Rostro-caudal, dorso-ventral (sagittal)
        _ = nice_heatmap(
            df,
            animals,
            x="Atlas_X",
            y="Atlas_Y",
            xlabel="Rostro-caudal (mm)",
            ylabel="Dorso-ventral (mm)",
            invertx=True,
            inverty=True,
            cmap="OrRd",
            rasterized=True,
            cbar=True,
        )

        # Medio-lateral, dorso-ventral (coronal)
        _ = nice_heatmap(
            df,
            animals,
            x="Atlas_Z",
            y="Atlas_Y",
            xlabel="Medio-lateral (mm)",
            ylabel="Dorso-ventral (mm)",
            inverty=True,
            invertx=True,
            cmap="OrRd",
            rasterized=True,
        )

    return fig_heatmap



"""io module, part of cuisto.

Contains loading and saving functions.

"""

import os

from brainglobe_atlasapi import BrainGlobeAtlas
from cuisto import utils
import orjson
import pandas as pd

JSON_KW = ["fibers", "axons", "fiber", "axon"]
CSV_KW = ["synapto", "synaptophysin", "syngfp", "boutons", "points"] + [
    "cells",
    "polygon",
    "polygons",
    "polygon",
    "cell",
]


def get_measurements_directory(wdir, animal: str, kind: str, segtype: str) -> str:
    """
    Get the directory with detections or annotations measurements for given animal ID.

    Parameters
    ----------
    wdir : str
        Base working directory.
    animal : str
        Animal ID.
    kind : str
        "annotation" or "detection".
    segtype : str
        Type of segmentation, eg. "synaptophysin".

    Returns
    -------
    directory : str
        Path to detections or annotations directory.

    """
    bdir = os.path.join(wdir, animal, animal.lower() + "_segmentation", segtype)

    if (kind == "detection") or (kind == "detections"):
        return os.path.join(bdir, "detections")
    elif (kind == "annotation") or (kind == "annotations"):
        return os.path.join(bdir, "annotations")
    else:
        raise ValueError(
            f"kind = '{kind}' not supported. Choose 'detection' or 'annotation'."
        )


def check_empty_file(filename: str, threshold: int = 1) -> bool:
    """
    Checks if a file is empty.

    Empty is defined as a file whose number of lines is lower than or equal to
    `threshold` (to allow for headers).

    Parameters
    ----------
    filename : str
        Full path to the file to check.
    threshold : int, optional
        If number of lines is lower than or equal to this value, it is considered as
        empty. Default is 1.

    Returns
    -------
    empty : bool
        True if the file is empty as defined above.

    """
    with open(filename, "rb") as fid:
        nlines = sum(1 for _ in fid)

    if nlines <= threshold:
        return True
    else:
        return False


def cat_csv_dir(directory, **kwargs) -> pd.DataFrame:
    """
    Scans a directory for csv files and concatenate them into a single DataFrame.

    Parameters
    ----------
    directory : str
        Path to the directory to scan.
    **kwargs : passed to pandas.read_csv()

    Returns
    -------
    df : pandas.DataFrame
        All CSV files concatenated in a single DataFrame.

    """
    return pd.concat(
        pd.read_csv(
            os.path.join(directory, filename),
            **kwargs,
        )
        for filename in os.listdir(directory)
        if (filename.endswith(".csv"))
        and not check_empty_file(os.path.join(directory, filename), threshold=1)
    )


def cat_json_dir(
    directory: str, hemisphere_names: dict, atlas: BrainGlobeAtlas
) -> pd.DataFrame:
    """
    Scans a directory for json files and concatenate them in a single DataFrame.

    The json files must be generated with 'pipelineImportExport.groovy" from a QuPath
    project.

    Parameters
    ----------
    directory : str
        Path to the directory to scan.
    hemisphere_names : dict
        Maps between hemisphere names in the json files ("Right" and "Left") to
        something else (eg. "Ipsi." and "Contra.").
    atlas : BrainGlobeAtlas
        Atlas to read regions from.

    Returns
    -------
    df : pd.DataFrame
        All JSON files concatenated in a single DataFrame.

    """
    # list files
    files_list = [
        os.path.join(directory, filename)
        for filename in os.listdir(directory)
        if (filename.endswith(".json"))
    ]

    data = []  # prepare list of DataFrame
    for filename in files_list:
        with open(filename, "rb") as fid:
            df = pd.DataFrame.from_dict(
                orjson.loads(fid.read())["paths"], orient="index"
            )
            df["Image"] = os.path.basename(filename).split("_detections")[0]
            data.append(df)

    df = (
        pd.concat(data)
        .explode(
            ["x", "y", "z", "hemisphere"]
        )  # get an entry for each point of segments
        .reset_index()
        .rename(
            columns=dict(
                x="Atlas_X",
                y="Atlas_Y",
                z="Atlas_Z",
                index="Object ID",
                classification="Classification",
            )
        )
        .set_index("Object ID")
    )

    # change hemisphere names
    df["hemisphere"] = df["hemisphere"].map(hemisphere_names)

    # add object type
    df["Object type"] = "Detection"

    # add brain regions
    df = utils.add_brain_region(df, atlas, col="Parent")

    return df


def cat_data_dir(directory: str, segtype: str, **kwargs) -> pd.DataFrame:
    """
    Wraps either cat_csv_dir() or cat_json_dir() depending on `segtype`.

    Parameters
    ----------
    directory : str
        Path to the directory to scan.
    segtype : str
        "synaptophysin" or "fibers".
    **kwargs : passed to cat_csv_dir() or cat_json_dir().

    Returns
    -------
    df : pd.DataFrame
        All files concatenated in a single DataFrame.

    """
    if segtype in CSV_KW:
        # remove kwargs for json
        kwargs.pop("hemisphere_names", None)
        kwargs.pop("atlas", None)
        return cat_csv_dir(directory, **kwargs)
    elif segtype in JSON_KW:
        kwargs = {k: kwargs[k] for k in ["hemisphere_names", "atlas"] if k in kwargs}
        return cat_json_dir(directory, **kwargs)
    else:
        raise ValueError(
            f"'{segtype}' not supported, unable to determine if CSV or JSON."
        )


def save_dfs(out_dir: str, filename, dfs: dict):
    """
    Save DataFrames to file.

    File format is inferred from file name extension.

    Parameters
    ----------
    out_dir : str
        Output directory.
    filename : _type_
        File name.
    dfs : dict
        DataFrames to save, as {identifier: df}. If HDF5 or xlsx, all df are saved in
        the same file, otherwise identifier is appended to the file name.

    """
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)

    basename, ext = os.path.splitext(filename)
    if ext in [".h5", ".hdf", ".hdf5"]:
        path = os.path.join(out_dir, filename)
        for identifier, df in dfs.items():
            df.to_hdf(path, key=identifier, mode="w")
    elif ext == ".xlsx":
        for identifier, df in dfs.items():
            df.to_excel(path, sheet_name=identifier)
    else:
        for identifier, df in dfs.items():
            path = os.path.join(out_dir, f"{basename}_{identifier}{ext}")
            if ext in [".pickle", ".pkl"]:
                df.to_pickle(path)
            elif ext == ".csv":
                df.to_csv(path)
            elif ext == ".tsv":
                df.to_csv(path, sep="\t")
            else:
                raise ValueError(f"{filename} has an unsupported extension.")


def load_dfs(
    filepath: str,
    fmt: str,
    identifiers: list[str] = [
        "df_regions",
        "df_coordinates",
        "df_distribution_ap",
        "df_distribution_dv",
        "df_distribution_ml",
    ],
):
    """
    Load DataFrames from file.

    If `fmt` is "h5" ("xslx"), identifiers are interpreted as h5 group identifier (sheet
    name, respectively).
    If `fmt` is "pickle", "csv" or "tsv", identifiers are appended to `filename`.
    Path to the file can't have a dot (".") in it.

    Parameters
    ----------
    filepath : str
        Full path to the file(s), without extension.
    fmt : {"h5", "csv", "pickle", "xlsx"}
        File(s) format.
    identifiers : list of str, optional
        List of identifiers to load from files. Defaults to the ones saved in
        cuisto.process.process_animals().

    Returns
    -------
    All requested DataFrames.

    """
    # ensure filename without extension
    base_path = os.path.splitext(filepath)[0]
    full_path = base_path + "." + fmt

    res = []
    if (fmt == "h5") or (fmt == "hdf") or (fmt == "hdf5"):
        for identifier in identifiers:
            res.append(pd.read_hdf(full_path, identifier))
    elif fmt == "xlsx":
        for identifier in identifiers:
            res.append(pd.read_excel(full_path, sheet_name=identifier))
    else:
        for identifier in identifiers:
            id_path = f"{base_path}_{identifier}.{fmt}"
            if (fmt == "pickle") or (fmt == "pkl"):
                res.append(pd.read_pickle(id_path))
            elif fmt == "csv":
                res.append(pd.read_csv(id_path))
            elif fmt == "tsv":
                res.append(pd.read_csv(id_path, sep="\t"))
            else:
                raise ValueError(f"{fmt} is not supported.")

    return res



"""compute module, part of cuisto.

Contains actual computation functions.

"""

import numpy as np
import pandas as pd

from cuisto.utils import get_starter_cells, select_hemisphere_channel


def get_regions_metrics(
    df_annotations: pd.DataFrame,
    object_type: str,
    channel_names: dict,
    meas_base_name: str,
    metrics_names: dict,
) -> pd.DataFrame:
    """
    Get a new DataFrame with cumulated axons segments length in each brain regions.

    This is the quantification per brain regions for fibers-like objects, eg. axons. The
    returned DataFrame has columns "cum. length µm", "cum. length mm", "density µm^-1",
    "density mm^-1", "coverage index".

    Parameters
    ----------
    df_annotations : pandas.DataFrame
        DataFrame with an entry for each brain regions, with columns "Area µm^2",
        "Name", "hemisphere", and "{object_type: channel} Length µm".
    object_type : str
        Object type (primary classification).
    channel_names : dict
        Map between original channel names to something else.
    meas_base_name : str
    metrics_names : dict

    Returns
    -------
    df_regions : pandas.DataFrame
        DataFrame with brain regions name, area and metrics.

    """
    # get columns names
    cols = df_annotations.columns
    # get columns with fibers lengths
    cols_colors = cols[
        cols.str.startswith(object_type) & cols.str.endswith(meas_base_name)
    ]
    # select relevant data
    cols_to_select = pd.Index(["Name", "hemisphere", "Area µm^2"]).append(cols_colors)
    # sum lengths and areas of each brain regions
    df_regions = (
        df_annotations[cols_to_select]
        .groupby(["Name", "hemisphere"])
        .sum()
        .reset_index()
    )

    # get measurement for both hemispheres (sum)
    df_both = df_annotations[cols_to_select].groupby(["Name"]).sum().reset_index()
    df_both["hemisphere"] = "both"
    df_regions = (
        pd.concat([df_regions, df_both], ignore_index=True)
        .sort_values(by="Name")
        .reset_index()
        .drop(columns="index")
    )

    # rename measurement columns to lower case
    df_regions = df_regions.rename(
        columns={
            k: k.replace(meas_base_name, meas_base_name.lower()) for k in cols_colors
        }
    )

    # update names
    meas_base_name = meas_base_name.lower()
    cols = df_regions.columns
    cols_colors = cols[
        cols.str.startswith(object_type) & cols.str.endswith(meas_base_name)
    ]

    # convert area in mm^2
    df_regions["Area mm^2"] = df_regions["Area µm^2"] / 1e6

    # prepare metrics
    if "µm" in meas_base_name:
        # fibers : convert to mm
        cols_to_convert = pd.Index([col for col in cols_colors if "µm" in col])
        df_regions[cols_to_convert.str.replace("µm", "mm")] = (
            df_regions[cols_to_convert] / 1000
        )
        metrics = [meas_base_name, meas_base_name.replace("µm", "mm")]
    else:
        # objects : count
        metrics = [meas_base_name]

    # density = measurement / area
    metric = metrics_names["density µm^-2"]
    df_regions[cols_colors.str.replace(meas_base_name, metric)] = df_regions[
        cols_colors
    ].divide(df_regions["Area µm^2"], axis=0)
    metrics.append(metric)
    metric = metrics_names["density mm^-2"]
    df_regions[cols_colors.str.replace(meas_base_name, metric)] = df_regions[
        cols_colors
    ].divide(df_regions["Area mm^2"], axis=0)
    metrics.append(metric)

    # coverage index = measurement² / area
    metric = metrics_names["coverage index"]
    df_regions[cols_colors.str.replace(meas_base_name, metric)] = (
        df_regions[cols_colors].pow(2).divide(df_regions["Area µm^2"], axis=0)
    )
    metrics.append(metric)

    # prepare relative metrics columns
    metric = metrics_names["relative measurement"]
    cols_rel_meas = cols_colors.str.replace(meas_base_name, metric)
    df_regions[cols_rel_meas] = np.nan
    metrics.append(metric)
    metric = metrics_names["relative density"]
    cols_dens = cols_colors.str.replace(meas_base_name, metrics_names["density mm^-2"])
    cols_rel_dens = cols_colors.str.replace(meas_base_name, metric)
    df_regions[cols_rel_dens] = np.nan
    metrics.append(metric)
    # relative metrics should be defined within each hemispheres (left, right, both)
    for hemisphere in df_regions["hemisphere"].unique():
        row_indexer = df_regions["hemisphere"] == hemisphere

        # relative measurement = measurement / total measurement
        df_regions.loc[row_indexer, cols_rel_meas] = (
            df_regions.loc[row_indexer, cols_colors]
            .divide(df_regions.loc[row_indexer, cols_colors].sum())
            .to_numpy()
        )

        # relative density = density / total density
        df_regions.loc[row_indexer, cols_rel_dens] = (
            df_regions.loc[
                row_indexer,
                cols_dens,
            ]
            .divide(df_regions.loc[row_indexer, cols_dens].sum())
            .to_numpy()
        )

    # collect channel names
    channels = (
        cols_colors.str.replace(object_type + ": ", "")
        .str.replace(" " + meas_base_name, "")
        .values.tolist()
    )
    # collect measurements columns names
    cols_metrics = df_regions.columns.difference(
        pd.Index(["Name", "hemisphere", "Area µm^2", "Area mm^2"])
    )
    for metric in metrics:
        cols_to_cat = [f"{object_type}: {cn} {metric}" for cn in channels]
        # make sure it's part of available metrics
        if not set(cols_to_cat) <= set(cols_metrics):
            raise ValueError(f"{cols_to_cat} not in DataFrame.")
        # group all colors in the same colors
        df_regions[metric] = df_regions[cols_to_cat].values.tolist()
        # remove original data
        df_regions = df_regions.drop(columns=cols_to_cat)

    # add a color tag, given their names in the configuration file
    df_regions["channel"] = len(df_regions) * [[channel_names[k] for k in channels]]
    metrics.append("channel")

    # explode the dataframe so that each color has an entry
    df_regions = df_regions.explode(metrics)

    return df_regions


def get_distribution(
    df: pd.DataFrame,
    col: str,
    hue: str,
    hue_filter: dict,
    per_commonnorm: bool,
    binlim: tuple | list,
    nbins=100,
) -> pd.DataFrame:
    """
    Computes distribution of objects.

    A global distribution using only `col` is computed, then it computes a distribution
    distinguishing values in the `hue` column. For the latter, it is possible to use a
    subset of the data ony, based on another column using `hue_filter`. This another
    column is determined with `hue`, if the latter is "hemisphere", then `hue_filter` is
    used in the "channel" color and vice-versa.
    `per_commonnorm` controls how they are normalized, either as a whole (True) or
    independantly (False).

    Use cases :
    (1) single-channel, two hemispheres : `col=x`, `hue=hemisphere`, `hue_filter=""`,
    `per_commonorm=True`. Computes a distribution for each hemisphere, the sum of the
    area of both is equal to 1.
    (2) three-channels, one hemisphere : `col=x`, hue=`channel`,
    `hue_filter="Ipsi.", per_commonnorm=False`. Computes a distribution for each channel
    only for points in the ipsilateral hemisphere. Each curve will have an area of 1.

    Parameters
    ----------
    df : pandas.DataFrame
    col : str
        Key in `df`, used to compute the distributions.
    hue : str
        Key in `df`. Criterion for additional distributions.
    hue_filter : str
        Further filtering for "per" distribution.
        - hue = channel : value is the name of one of the hemisphere
        - hue = hemisphere : value can be the name of a channel, a list of such or "all"
    per_commonnorm : bool
        Use common normalization for all hues (per argument).
    binlim : list or tuple
        First bin left edge and last bin right edge.
    nbins : int, optional
        Number of bins. Default is 100.

    Returns
    -------
    df_distribution : pandas.DataFrame
        DataFrame with `bins`, `distribution`, `count` and their per-hemisphere or
        per-channel variants.

    """

    # - Preparation
    bin_edges = np.linspace(*binlim, nbins + 1)  # create bins
    df_distribution = []  # prepare list of distributions

    # - Both hemispheres, all channels
    # get raw count per bins (histogram)
    count, bin_edges = np.histogram(df[col], bin_edges)
    # get normalized count (pdf)
    distribution, _ = np.histogram(df[col], bin_edges, density=True)
    # get bin centers rather than edges to plot them
    bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2

    # make a DataFrame out of that
    df_distribution.append(
        pd.DataFrame(
            {
                "bins": bin_centers,
                "distribution": distribution,
                "count": count,
                "hemisphere": "both",
                "channel": "all",
                "axis": col,  # keep track of what col. was used
            }
        )
    )

    # - Per additional criterion
    # select data
    df_sub = select_hemisphere_channel(df, hue, hue_filter, False)
    hue_values = df[hue].unique()  # get grouping values
    # total number of datapoints in the subset used for additional distribution
    length_total = len(df_sub)

    for value in hue_values:
        # select part and coordinates
        df_part = df_sub.loc[df_sub[hue] == value, col]

        # get raw count per bins (histogram)
        count, bin_edges = np.histogram(df_part, bin_edges)
        # get normalized count (pdf)
        distribution, _ = np.histogram(df_part, bin_edges, density=True)

        if per_commonnorm:
            # re-normalize so that the sum of areas of all sub-parts is 1
            length_part = len(df_part)  # number of datapoints in that hemisphere
            distribution *= length_part / length_total

        # get bin centers rather than edges to plot them
        bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2

        # make a DataFrame out of that
        df_distribution.append(
            pd.DataFrame(
                {
                    "bins": bin_centers,
                    "distribution": distribution,
                    "count": count,
                    hue: value,
                    "channel" if hue == "hemisphere" else "hemisphere": hue_filter,
                    "axis": col,  # keep track of what col. was used
                }
            )
        )

    return pd.concat(df_distribution)


def normalize_starter_cells(
    df: pd.DataFrame, cols: list[str], animal: str, info_file: str, channel_names: dict
) -> pd.DataFrame:
    """
    Normalize data by the number of starter cells.

    Parameters
    ----------
    df : pd.DataFrame
        Contains the data to be normalized.
    cols : list-like
        Columns to divide by the number of starter cells.
    animal : str
        Animal ID to parse the number of starter cells.
    info_file : str
        Full path to the TOML file with informations.
    channel_names : dict
        Map between original channel names to something else.

    Returns
    -------
    pd.DataFrame
        Same `df` with normalized count.

    """
    for channel in df["channel"].unique():
        # inverse mapping channel colors : names
        reverse_channels = {v: k for k, v in channel_names.items()}
        nstarters = get_starter_cells(animal, reverse_channels[channel], info_file)

        for col in cols:
            df.loc[df["channel"] == channel, col] = (
                df.loc[df["channel"] == channel, col] / nstarters
            )

    return df



"""seg module, part of cuisto.

Functions for segmentating probability map stored as an image.

"""

import uuid

import geojson
import numpy as np
import pandas as pd
import shapely
import tifffile
from skan import Skeleton, summarize
from skimage import measure, morphology


def get_pixelsize(image_name: str) -> float:
    """
    Get pixel size recorded in `image_name` TIFF metadata.

    Parameters
    ----------
    image_name : str
        Full path to image.

    Returns
    -------
    pixelsize : float
        Pixel size in microns.

    """

    with tifffile.TiffFile(image_name) as tif:
        # XResolution is a tuple, numerator, denomitor. The inverse is the pixel size
        return (
            tif.pages[0].tags["XResolution"].value[1]
            / tif.pages[0].tags["XResolution"].value[0]
        )


def convert_to_pixels(filters, pixelsize):
    """
    Convert some values in `filters` in pixels.

    Parameters
    ----------
    filters : dict
        Must contain the keys used below.
    pixelsize : float
        Pixel size in microns.

    Returns
    -------
    filters : dict
        Same as input, with values in pixels.

    """

    filters["area_low"] = filters["area_low"] / pixelsize**2
    filters["area_high"] = filters["area_high"] / pixelsize**2
    filters["length_low"] = filters["length_low"] / pixelsize
    filters["dist_thresh"] = int(filters["dist_thresh"] / pixelsize)

    return filters


def pad_image(img: np.ndarray, finalsize: tuple | list) -> np.ndarray:
    """
    Pad image with zeroes to match expected final size.

    Parameters
    ----------
    img : ndarray
    finalsize : tuple or list
        nrows, ncolumns

    Returns
    -------
    imgpad : ndarray
        img with black borders.

    """

    final_h = finalsize[0]  # requested number of rows (height)
    final_w = finalsize[1]  # requested number of columns (width)
    original_h = img.shape[0]  # input number of rows
    original_w = img.shape[1]  # input number of columns

    a = (final_h - original_h) // 2  # vertical padding before
    aa = final_h - a - original_h  # vertical padding after
    b = (final_w - original_w) // 2  # horizontal padding before
    bb = final_w - b - original_w  # horizontal padding after

    return np.pad(img, pad_width=((a, aa), (b, bb)), mode="constant")


def erode_mask(mask: np.ndarray, edge_dist: float) -> np.ndarray:
    """
    Erode the mask outline so that is is `edge_dist` smaller from the border.

    This allows discarding the edges.

    Parameters
    ----------
    mask : ndarray
    edge_dist : float
        Distance to edges, in pixels.

    Returns
    -------
    eroded_mask : ndarray of bool

    """

    if edge_dist % 2 == 0:
        edge_dist += 1  # decomposition requires even number

    footprint = morphology.square(edge_dist, decomposition="sequence")

    return mask * morphology.binary_erosion(mask, footprint=footprint)


def get_image_skeleton(img: np.ndarray, minsize=0) -> np.ndarray:
    """
    Get the image skeleton.

    Computes the image skeleton and removes objects smaller than `minsize`.

    Parameters
    ----------
    img : ndarray of bool
    minsize : number, optional
        Min. size the object can have, as a number of pixels. Default is 0.

    Returns
    -------
    skel : ndarray of bool
        Binary image with 1-pixel wide skeleton.

    """

    skel = morphology.skeletonize(img)

    return morphology.remove_small_objects(skel, min_size=minsize, connectivity=2)


def get_collection_from_skel(
    skeleton: Skeleton, properties: dict, rescale_factor: float = 1.0, offset=0.5
) -> geojson.FeatureCollection:
    """
    Get the coordinates of each skeleton path as a GeoJSON Features in a
    FeatureCollection.
    `properties` is a dictionnary with QuPath properties of each detections.

    Parameters
    ----------
    skeleton : skan.Skeleton
    properties : dict
        QuPatj objects' properties.
    rescale_factor : float
        Rescale output coordinates by this factor.
    offset : float
        Shift coordinates by this amount, typically to get pixel centers or edges.
        Default is 0.5.

    Returns
    -------
    collection : geojson.FeatureCollection
        A FeatureCollection ready to be written as geojson.

    """

    branch_data = summarize(skeleton, separator="_")

    collection = []
    for ind in range(skeleton.n_paths):
        prop = properties.copy()
        prop["measurements"] = {"skeleton_id": int(branch_data.loc[ind, "skeleton_id"])}
        collection.append(
            geojson.Feature(
                geometry=shapely.LineString(
                    (skeleton.path_coordinates(ind)[:, ::-1] + offset) * rescale_factor
                ),  # shape object
                properties=prop,  # object properties
                id=str(uuid.uuid4()),  # object uuid
            )
        )

    return geojson.FeatureCollection(collection)


def get_collection_from_poly(
    contours: list, properties: dict, rescale_factor: float = 1.0, offset: float = 0.5
) -> geojson.FeatureCollection:
    """
    Gather coordinates in the list and put them in GeoJSON format as Polygons.

    An entry in `contours` must define a closed polygon. `properties` is a dictionnary
    with QuPath properties of each detections.

    Parameters
    ----------
    contours : list
    properties : dict
        QuPatj objects' properties.
    rescale_factor : float
        Rescale output coordinates by this factor.
    offset : float
        Shift coordinates by this amount, typically to get pixel centers or edges.
        Default is 0.5.

    Returns
    -------
    collection : geojson.FeatureCollection
        A FeatureCollection ready to be written as geojson.

    """
    collection = [
        geojson.Feature(
            geometry=shapely.Polygon(
                np.fliplr((contour + offset) * rescale_factor)
            ),  # shape object
            properties=properties,  # object properties
            id=str(uuid.uuid4()),  # object uuid
        )
        for contour in contours
    ]

    return geojson.FeatureCollection(collection)


def get_collection_from_points(
    coords: list, properties: dict, rescale_factor: float = 1.0, offset: float = 0.5
) -> geojson.FeatureCollection:
    """
    Gather coordinates from `coords` and put them in GeoJSON format.

    An entry in `coords` are pairs of (x, y) coordinates defining the point.
    `properties` is a dictionnary with QuPath properties of each detections.

    Parameters
    ----------
    coords : list
    properties : dict
    rescale_factor : float
        Rescale output coordinates by this factor.

    Returns
    -------
    collection : geojson.FeatureCollection

    """

    collection = [
        geojson.Feature(
            geometry=shapely.Point(
                np.flip((coord + offset) * rescale_factor)
            ),  # shape object
            properties=properties,  # object properties
            id=str(uuid.uuid4()),  # object uuid
        )
        for coord in coords
    ]

    return geojson.FeatureCollection(collection)


def segment_lines(
    img: np.ndarray, geojson_props: dict, minsize=0.0, rescale_factor=1.0
) -> geojson.FeatureCollection:
    """
    Wraps skeleton analysis to get paths coordinates.

    Parameters
    ----------
    img : ndarray of bool
        Binary image to segment as lines.
    geojson_props : dict
        GeoJSON properties of objects.
    minsize : float
        Minimum size in pixels for an object.
    rescale_factor : float
        Rescale output coordinates by this factor.

    Returns
    -------
    collection : geojson.FeatureCollection
        A FeatureCollection ready to be written as geojson.

    """

    skel = get_image_skeleton(img, minsize=minsize)

    # get paths coordinates as FeatureCollection
    skeleton = Skeleton(skel, keep_images=False)
    return get_collection_from_skel(
        skeleton, geojson_props, rescale_factor=rescale_factor
    )


def segment_polygons(
    img: np.ndarray,
    geojson_props: dict,
    area_min: float = 0.0,
    area_max: float = np.inf,
    ecc_min: float = 0.0,
    ecc_max: float = 1.0,
    rescale_factor: float = 1.0,
) -> geojson.FeatureCollection:
    """
    Polygon segmentation.

    Parameters
    ----------
    img : ndarray of bool
        Binary image to segment as polygons.
    geojson_props : dict
        GeoJSON properties of objects.
    area_min, area_max : float
        Minimum and maximum area in pixels for an object.
    ecc_min, ecc_max : float
        Minimum and maximum eccentricity for an object.
    rescale_factor: float
        Rescale output coordinates by this factor.

    Returns
    -------
    collection : geojson.FeatureCollection
        A FeatureCollection ready to be written as geojson.

    """

    label_image = measure.label(img)

    # get objects properties
    stats = pd.DataFrame(
        measure.regionprops_table(
            label_image, properties=("label", "area", "eccentricity")
        )
    )

    # remove objects not matching filters
    toremove = stats[
        (stats["area"] < area_min)
        | (stats["area"] > area_max)
        | (stats["eccentricity"] < ecc_min)
        | (stats["eccentricity"] > ecc_max)
    ]

    label_image[np.isin(label_image, toremove["label"])] = 0

    # find objects countours
    label_image = label_image > 0
    contours = measure.find_contours(label_image)

    return get_collection_from_poly(
        contours, geojson_props, rescale_factor=rescale_factor
    )


def segment_points(
    img: np.ndarray,
    geojson_props: dict,
    area_min: float = 0.0,
    area_max: float = np.inf,
    ecc_min: float = 0,
    ecc_max: float = 1,
    dist_thresh: float = 0,
    rescale_factor: float = 1,
) -> geojson.FeatureCollection:
    """
    Point segmentation.

    First, segment polygons to apply shape filters, then extract their centroids,
    and remove isolated points as defined by `dist_thresh`.

    Parameters
    ----------
    img : ndarray of bool
        Binary image to segment as points.
    geojson_props : dict
        GeoJSON properties of objects.
    area_min, area_max : float
        Minimum and maximum area in pixels for an object.
    ecc_min, ecc_max : float
        Minimum and maximum eccentricity for an object.
    dist_thresh : float
        Maximal distance in pixels between objects before considering them as isolated and remove them.
        0 disables it.
    rescale_factor : float
        Rescale output coordinates by this factor.

    Returns
    -------
    collection : geojson.FeatureCollection
        A FeatureCollection ready to be written as geojson.

    """

    # get objects properties
    stats = pd.DataFrame(
        measure.regionprops_table(
            measure.label(img), properties=("label", "area", "eccentricity", "centroid")
        )
    )

    # keep objects matching filters
    stats = stats[
        (stats["area"] >= area_min)
        & (stats["area"] <= area_max)
        & (stats["eccentricity"] >= ecc_min)
        & (stats["eccentricity"] <= ecc_max)
    ]

    # create an image from centroids only
    stats["centroid-0"] = stats["centroid-0"].astype(int)
    stats["centroid-1"] = stats["centroid-1"].astype(int)
    bw = np.zeros(img.shape, dtype=bool)
    bw[stats["centroid-0"], stats["centroid-1"]] = True

    # filter isolated objects
    if dist_thresh:
        # dilation of points
        if dist_thresh % 2 == 0:
            dist_thresh += 1  # decomposition requires even number

        footprint = morphology.square(int(dist_thresh), decomposition="sequence")
        dilated = measure.label(morphology.binary_dilation(bw, footprint=footprint))
        stats = pd.DataFrame(
            measure.regionprops_table(dilated, properties=("label", "area"))
        )

        # objects that did not merge are alone
        toremove = stats[(stats["area"] <= dist_thresh**2)]
        dilated[np.isin(dilated, toremove["label"])] = 0  # remove them

        # apply mask
        bw = bw * dilated

    # get points coordinates
    coords = np.argwhere(bw)

    return get_collection_from_points(
        coords, geojson_props, rescale_factor=rescale_factor
    )



"""cuisto package.
Perform quantification of objects in registered and segmented histological slices.
"""

from . import compute, display, io, process, utils, seg
from .config import Config

__all__ = ["Config", "compute", "display", "io", "process", "utils", "seg"]



"""config module, part of cuisto.

Contains the Config class.

"""

import tomllib
import warnings

from brainglobe_atlasapi import BrainGlobeAtlas
from cuisto import utils


class Config:
    """
    The configuration class.

    Reads input configuration file and provides its constant.

    Parameters
    ----------
    config_file : str
        Full path to the configuration file to load.

    Returns
    -------
    cfg : Config object.

    """

    def __init__(self, config_file):
        """Constructor."""
        with open(config_file, "rb") as fid:
            cfg = tomllib.load(fid)

            for key in cfg:
                setattr(self, key, cfg[key])

        self.config_file = config_file
        if self.atlas["name"]:
            self.bg_atlas = BrainGlobeAtlas(self.atlas["name"], check_latest=False)
        else:
            self.bg_atlas = None

        self.get_blacklist()
        self.get_leaves_list()

    def get_blacklist(self):
        """Wraps cuisto.utils.get_blacklist."""

        self.atlas["blacklist"] = utils.get_blacklist(
            self.files["blacklist"], self.bg_atlas
        )

    def get_leaves_list(self):
        """Wraps utils.get_leaves_list."""

        self.atlas["leaveslist"] = utils.get_leaves_list(self.bg_atlas)

    def get_injection_sites(self, animals: list[str]) -> dict:
        """
        Get list of injection sites coordinates for each animals, for each channels.

        Parameters
        ----------
        animals : list of str
            List of animals.

        Returns
        -------
        injection_sites : dict
            {"x": {channel0: [x]}, "y": {channel1: [y]}}

        """
        injection_sites = {
            axis: {channel: [] for channel in self.channels["names"].keys()}
            for axis in ["x", "y", "z"]
        }

        for animal in animals:
            for channel in self.channels["names"].keys():
                injx, injy, injz = utils.get_injection_site(
                    animal,
                    self.files["infos"],
                    channel,
                    stereo=self.distributions["stereo"],
                )
                if injx is not None:
                    injection_sites["x"][channel].append(injx)
                if injy is not None:
                    injection_sites["y"][channel].append(injy)
                if injz is not None:
                    injection_sites["z"][channel].append(injz)

        return injection_sites

    def get_hue_palette(self, mode: str) -> dict:
        """
        Get color palette given hue.

        Maps hue to colors in channels or hemispheres.

        Parameters
        ----------
        mode : {"hemisphere", "channel"}

        Returns
        -------
        palette : dict
            Maps a hue level to a color, usable in seaborn.

        """
        params = getattr(self, mode)

        if params["hue"] == "channel":
            # replace channels by their new names
            palette = {
                self.channels["names"][k]: v for k, v in self.channels["colors"].items()
            }
        elif params["hue"] == "hemisphere":
            # replace hemispheres by their new names
            palette = {
                self.hemispheres["names"][k]: v
                for k, v in self.hemispheres["colors"].items()
            }
        else:
            palette = None
            warnings.warn(f"hue={self.regions["display"]["hue"]} not supported.")

        return palette



"""
Run a QuPath script in a parallel loop.

Requires paquo to get the list of images.

"""

import os
import multiprocessing
import subprocess

from functools import partial
from paquo.projects import QuPathProject
from tqdm import tqdm

NTHREADS = int(os.cpu_count() / 2)


def get_images_list(qupath_project: str) -> list[str]:
    """
    Get the QuPath project images list.

    Parameters
    ----------
    qupath_project : str
        Full path the project.qpproj file.

    Returns
    -------
    images_list : list of str
        List of images names.

    """

    qp = QuPathProject(qupath_project, "r")
    return [img.image_name for img in qp.images]


def run_script(
    qupath_exe: str,
    script_path: str,
    qupath_project: str,
    quiet: bool,
    save: bool,
    imgname: str,
):
    """
    Run groovy script with QuPath in console mode on selected image.

    Parameters
    ----------
    qupath_exe : str
        Full path to the QuPath console exe.
    script_path : str
        Full path to the groovy script to be executed.
    qupath_project : str
        Full path to the QuPath .qproj file.
    quiet : bool
        Suppress console verbose. Default is False.
    save : bool, optional
        Update data files for each image in the project. Default is False.
    imgname : str
        Image name in the QuPath project.
    """

    if quiet:
        stdout = subprocess.DEVNULL
        log = "OFF"
    else:
        stdout = None
        log = "INFO"

    if save:
        save_arg = "-s"
    else:
        save_arg = ""

    cmd = [
        qupath_exe,
        "--quiet",
        "--log",
        log,
        "script",
        script_path,
        "-p",
        qupath_project,
        "-i",
        imgname,
        save_arg,
    ]
    subprocess.run(cmd, shell=True, stdout=stdout)


def multirun(
    qupath_exe: str,
    script_path: str,
    qupath_project: str,
    quiet: bool = False,
    save: bool = False,
    nthreads: int = NTHREADS,
    exclude_list: list[str] = [],
):
    """
    Run groovy script on QuPath project images, in multi-thread.

    Parameters
    ----------
    qupath_exe : str
        Full path to the QuPath console exe.
    script_path : str
        Full path to the groovy script to be executed.
    qupath_project : str
        Full path to the QuPath .qproj file.
    quiet : bool, optional
        Suppress console verbose. Default is False.
    save : bool, optional
        Update data files for each image in the project. Default is False.
    nthreads : int, optional
        Number of threads. Default is half the number of cores of the computer.
    exclude_list : list[str], optional
        List of images name to NOT execute the script on. Default is empty list.

    """

    # get list of images
    imglist = get_images_list(qupath_project)
    # exclude specified images
    imglist = [imgname for imgname in imglist if imgname not in exclude_list]

    # multithreaded call to the script via QuPath console
    with multiprocessing.Pool(nthreads) as pool:
        # prepare the lazy map
        runner = tqdm(
            pool.imap(
                partial(run_script, qupath_exe, script_path, qupath_project, quiet, save),
                imglist,
            ),
            total=len(imglist),
        )

        # execute
        tuple(runner)
